"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const tsutils_1 = require("tsutils");
const typescript_1 = __importDefault(require("typescript"));
/**
 * An ESLint rule to ban unsafe array/object destructuring, which is not well-typed in TypeScript.
 */
const noArrayDestructuring = {
    meta: {
        type: "problem",
        docs: {
            category: "Possible Errors",
            description: "Array and object destructuring are not type-safe in TypeScript.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringGeneric: "This destructuring is not type-safe in TypeScript.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        return {
            // eslint-disable-next-line functional/no-return-void
            ArrayPattern: (node) => {
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const type = checker.getTypeAtLocation(tsNode);
                // eslint-disable-next-line functional/no-conditional-statement
                if (tsutils_1.isTupleTypeReference(type) &&
                    node.elements.length <= type.target.minLength) {
                    // Allow tuples (whether partial or not) if the indexes
                    // we're destructuring are all know to be safe at compile time.
                    return;
                }
                const numberIndexType = type.getNumberIndexType();
                const typeParts = numberIndexType !== undefined
                    ? numberIndexType.isUnion()
                        ? numberIndexType.types
                        : [numberIndexType]
                    : [];
                // eslint-disable-next-line functional/no-conditional-statement
                if (typeParts.find((t) => t.flags & typescript_1.default.TypeFlags.Undefined) !== undefined) {
                    // Allow destructuring if undefined is already in the array type.
                    return;
                }
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            },
            // eslint-disable-next-line functional/no-return-void
            ObjectPattern: (node) => {
                const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
                const type = checker.getTypeAtLocation(tsNode);
                // True if every element we're destructuring is known to exist at compile time (or is the rest element).
                const isSafe = node.properties.every((prop) => prop.type === experimental_utils_1.AST_NODE_TYPES.RestElement ||
                    (prop.key.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                        type.getProperty(prop.key.name) !== undefined));
                // eslint-disable-next-line functional/no-conditional-statement
                if (isSafe) {
                    return;
                }
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            },
        };
    },
};
exports.default = noArrayDestructuring;
